/*         PMODEM.C
 
   This program is used to transfer files from a remote packet user
   to and from the packet host. Text files have the CR converted into
   CR, LF for compatability with non-packet circuits. (The packet Bios
   strips LF's during output to the packet channel, so this is transparent
   to the packet user)
 
   Written by Skip Hansen WB6YMH-0  8/84
   Modified by Tom Cain WB8OUE-0 5/85 for IBM PC type junk
 
   copyleft (c) 1984, all wrongs reserved
   permission given for free copying, provided credit is
   given to the guilty. Please don't sell it, charge for
   copying, etc...
 
   This program has two possible modes of operation, Binary and Text.
   The Text mode is intended to be used by users with "text capture"
   terminal programs without protocol. The requested file is transmitted/
   received with CR converted to CRLF and visa versa. The end of the file
   is indicated by a CP/M EOF mark (1AH or Ctrl-Z). The Binary mode is
   used with a complementary program at the users end.
 
   DLE-ENQ
 
   If a data byte of DLE occurs in the data stream it is encoded as
   DLE-DLE.  When the user is sending a binary file to the packet host
   he may abort the file transfer at any time by the sequence DLE-CAN.
 
  DLE = 0x10    STX = 0x02      ETX = 0x03      ENQ = 0x05
        Ctrl-P        Ctrl-B          Ctrl-C          Ctrl-E
 
   The CRC polynominal is X^16 + X^12 + X^5 + 1
 
        Files of this program:
        PMODEM.C        This file.
 
        To compile and link using C-Systems compiler and library and IBM linker:
        cc pmodem.c
        link pmodem+setint0c,,,slib
 
*/
 
#include	<stdio.h>

#define LF	10
#define CR	13
#define DLE	0x10
#define CAN	0x18
#define ETX	0x3
#define STX	0x2
#define CPMEOF	0x1A

#define SPOOL_SIZE	8192	/* incoming data spool size */
#define WORD_8N1	0x03

#define PORT_LCR	0x3fb	/* pc com1: port addresses */
#define PORT_DLM	0x3f9
#define PORT_DLL	0x3f8
#define PORT_IIR	0x3fa
#define PORT_IER	0x3f9
#define PORT_MCR	0x3fc
#define PORT_RHR	0x3f8
#define PORT_THR	0x3f8
#define PORT_LSR	0x3fd
#define PORT_MSR	0x3fe

#define COM1_INT	0x10	/* com1: interuppt mask bit */

char	head_p[SPOOL_SIZE] ;	/* incoming data spool */
char	*tail_p, *read_p, *write_p ;
int	data_avail ;
 
FILE	*file1 ;
 
main(arg_ct, p_argv)

  char **p_argv;
  int arg_ct;
 
{

  int	c,x;
  int	binary,done,header,crcdata;
  int	baud_divisor ;
  int	com1int() ;

header=FALSE;
done=FALSE;
printf("PMODEM v1.1 05/22/85\n");
if(arg_ct != 4) usage();        /* must be 4 parameters or error */

switch (atoi(p_argv[1])) {	/*  set the com port speed variable */
  case 300:
       baud_divisor = 0x0180 ;
       break ;
  case 600:
       baud_divisor = 0x00c0 ;
       break ;
  case 1200:
       baud_divisor = 0x0060 ;
       break ;
  case 2400:
       baud_divisor = 0x0030 ;
       break ;
  case 4800:
       baud_divisor = 0x0018 ;
       break ;
  case 9600:
       baud_divisor = 0x000c ;
       break ;
  default:
      usage() ;
}  /* switch() */

binary=index(p_argv[2],"B") != ERROR;   /* set transparent flag if specified */
iportb(baud_divisor);                   /* initalize the 820 printer (TNC) port */
unmask_ints(COM1_INT) ;

if(index(p_argv[2],"R") != ERROR){
        if(open(p_argv[3],0) != ERROR){
                printf("The file \"%s\" already exists, please",p_argv[3]);
                puts(" use another name.\n");
                kill_232() ;
                exit(0);
        }
        file1 = fopen(p_argv[3],"w") ;
        printf("Ready to receive \"%s\" ",p_argv[3]);
        if(binary){
                printf("in binary mode.\n");
                while(!done){
                        while(!header){                 /* wait for DLE STX */
                                while(printf("%h ",getncc() != DLE));
                                if(getncc() == STX){
 
                                }
                        }
                        crcdata=updcrc(STX,updcrc(DLE,0));
                        c=getncc();
                        crcdata=updcrc(c,crcdata);
                        if(c == DLE){
                                x=getncc();
                                crcdata=updcrc(x,crcdata);
                                if(x == DLE)
                                        putc(c,file1);
                                else if(x == CAN) {
                                        unlink(p_argv[3]);
                                        puts("File transfer canceled at");
                                        puts(" senders request.\n");
                                        kill_232() ;
                                        exit(0);
                                }
                                else if(x == ETX) {
                                        crcdata=updcrc(getncc(),crcdata);
                                        if(updcrc(getncc(),crcdata) != 0)
 
                                        done=TRUE;
                                }
                        }
                        else putc(c,file1);
                }
        }
        else {
                printf("in text mode.\n");
                while((c=getncc()) != CPMEOF){
                        if(c != LF) {
                                putc(c,file1);
                                putchar(c);
                        }
                        if(c == CR){
                                putc(LF,file1); /* convert CR to CR, LF      */
                                putchar(LF);
                        }
                }
                putc(CPMEOF,file1);     /* append EOF mark */
        }
        if(fclose(file1) == ERROR){
                puts("Error closing file %s\n",p_argv[3]);
                kill_232() ;
                exit(0);
        }
}
else if(index(p_argv[2],"S") != ERROR){
        if((file1 = fopen(p_argv[3],"r")) == ERROR) {
                printf("File \"%s\" not found\n",p_argv[3]);
                kill_232() ;
                exit(0);
        }
        if(binary){
                printf("Sending file \"%s\" in binary mode.\n",p_argv[3]);
                outncc(DLE);    /* send header */
                outncc(STX);
                crcdata=updcrc(STX,updcrc(DLE,0));
                while((c=getc(file1)) != ERROR){
                        if(c == DLE){
                                outncc(DLE);
                                crcdata=updcrc(DLE,crcdata);
                        }
                        outncc(c);
                        crcdata=updcrc(c,crcdata);
                }
                outncc(DLE);
                outncc(ETX);
                outncc(crcdata>>8);
                outncc(crcdata);
                kill_232() ;
                exit(0);
        }
        else {
                while((c=getc(file1)) != ERROR){
                        outncc(c);
                        if(c == CPMEOF) {
                                kill_232() ;
                                exit(0);
                        }
                        putchar(c);
                }
                outncc(CPMEOF);
                kill_232() ;
                exit(0);
        }
}
else usage();
}  /* main() */


/***********************************************************************
;* CRCSUBS (Cyclic Redundancy Code Subroutines) version 1.20            *
;* 8080 Mnemonics, translated to 8086 Mnemonics by T. Cain 5/22/85      *
;*                                                                      *
;*      This subroutine will compute and check a true 16-bit            *
;*      Cyclic Redundancy Code for a message of arbitrary length.       *
;*                                                                      *
;*      The  use  of this scheme will guarantee detection of all        *
;*      single and double bit errors, all  errors  with  an  odd        *
;*      number  of  error bits, all burst errors of length 16 or        *
;*      less, 99.9969% of all 17-bit error bursts, and  99.9984%        *
;*      of  all  possible  longer  error bursts.  (Ref: Computer        *
;*      Networks, Andrew S.  Tanenbaum, Prentiss-Hall, 1981)            *
;*                                                                      *
;*                                                                      *
;************************************************************************
;*                                                                      *
;*      From: CRCSUB12.ASM                                              *
;*      Designed & coded by Paul Hansknecht, June 13, 1981              *
;*                                                                      *
;*                                                                      *
;*      Copyright (c) 1981, Carpenter Associates                        *
;*                          Box 451                                     *
;*                          Bloomfield Hills, MI 48013                  *
;*                          313/855-3074                                *
;*                                                                      *
;*      This program may be freely reproduced for non-profit use.       *
;*                                                                      *
;************************************************************************
;
;       unsigned updcrc(char, oldcrc)
;
;       At start of packet, oldcrc is set to initial value
;               oldcrc=0;
;
;       crc is accumulated by:
;               oldcrc=updcrc(char, oldcrc);
;
;       at end of packet,
;               oldcrc=updcrc(0,updcrc(0,oldcrc));
;               send(oldcrc>>8); send(oldcrc);
;
;       on receive, the return value of updcrc is checked after the
;       last call (with the second CRC byte); 0 indicates no error detected
;
 

*/
 
updcrc(c, oldcrc)

                        /* DON'T MOVE OR CHANGE THESE PARAMETERS! */         
  char	c ;		/* char to add to running crc */
  int	oldcrc ;	/* running crc */

{
  #asm
   mov  cl, [bp+8]	; char
   mov  bx, [bp+10]	; oldcrc
   mov  ch, 8		; shift counter
updloop:
   rol  cl, 1
   rcl  bl, 1
   rcl  bh, 1
   jnc  skipit
   xor  bx, 1021h
skipit:
   dec  ch
   jnz  updloop
   mov  [bp+10], bx	; return crc
  #endasm

  return(oldcrc) ;

}  /* updcrc() */


/**************************************************************************

   initialize com1: for tnc i/o

 **************************************************************************/

iportb(baud_divisor)		/* init the com1: port */

  int baud_divisor ;

{
  setint0c (com1int) ;		/* handle com1: int. */

  tail_p = &head_p[SPOOL_SIZE - 1] ;	/* init. the spool */
  read_p = head_p ;
  write_p = head_p ;
  data_avail = FALSE ;		/* no data yet */

  /* set up the port itself */
  outbyt(PORT_LCR, 0x80 | WORD_8N1) ;	/* access the divisor regs. */
  outbyt(PORT_DLM, baud_divisor / 256) ; /* set the baud rate */
  outbyt(PORT_DLL, baud_divisor % 256) ;
  outbyt(PORT_LCR, WORD_8N1) ;		/* 8 bits, 1 stop, no parity */
  outbyt(PORT_IIR, 1) ;			/* port reset */

  outbyt(PORT_IER, 1) ;			/* enable recvd data int. */
  outbyt(PORT_MCR, 0xb) ;		/* set dtr, rts, enable ints. */
  
}  /* ioport() */


/**************************************************************************

   com1: interuppt handler

   interuppts only on incoming data

 **************************************************************************/

com1int()		/* com1: interuppt handler */

{

  char	   dummy_char ;

/* This routine saves the current contents of the ds and es registers.
   This is necessary since this interrupt routine could interrupt any
   running program, including the PC-DOS routines.  Also, the ds and
   registers would not be set up properly for THIS routine, so they
   are initialized properly. */

  #asm
    push  ds
    push  es
    mov   ax, dgrp
    mov   ds, ax
    push  ds
    pop   es
  #endasm

  data_avail = TRUE ;

  #asm
   sti
  #endasm

/* Get the character that caused the interrupt and place it into the
   character spool.  If the spool is full, do not save it. */

  if (write_p == tail_p) {	       /* at the physical end of buffer */
    if (read_p != head_p) {            /* buffer not full */
      *write_p = inbyt (PORT_RHR) ;    /* store char in last phys position */
       write_p = head_p ;	       /* move write ptr to phys beginning */
       }
    else {     /* buffer is full */
      dummy_char = inbyt (PORT_RHR) ;
      kill_232() ;
      fprintf(stderr,"pmodem: input serial buffer blew\n") ;
      exit(0) ;
     }
    }
  else {				 /* not at physical end of buffer */
    if (write_p + 1 != read_p) {	 /* buffer not full */
      *write_p++ = inbyt (PORT_RHR) ;    /* store char in next phys position */
      }
    else {     /* buffer is full */
      dummy_char = inbyt (PORT_RHR) ;
      kill_232() ;
      fprintf(stderr,"pmodem : input serial buffer blew\n") ;
      exit(0) ;                      
     }
    }

  #asm			  /* send non-specific EOI */
    cli
    pop  es
    pop  ds
    mov  al, 20h
    out 20h, al
  #endasm

}    /* com1int() */




/****************************************************************************
 Procedure kill_232

 De-programs the RS-232 interface.

****************************************************************************/

kill_232 ()

{

  mask_ints (COM1_INT);			         /* disable 8259 interrupt */

  outbyt (PORT_IER, 0) ;			 /* Disable data avail int */

  }    /* kill_232 */


/****************************************************************************
 Procedure mask_ints

 Disables specified interrupt by reseting the level enable bit

****************************************************************************/

mask_ints(level)

  int level;
{
  char mask_bits;

  mask_bits = inbyt (0x21) ;
  mask_bits = mask_bits | level;
  outbyt (0x21,mask_bits);

}  /* mask_ints */


/****************************************************************************
 Procedure unmask_ints

 Enables specified interrupt by reseting the level bit in the mask

****************************************************************************/

unmask_ints(level)

  int level;
{
  char mask_bits ;  

  mask_bits = inbyt (0x21) ;
  mask_bits = mask_bits & ( level ^ -1 ) ;
  outbyt (0x21, mask_bits);

}



/****************************************************************************
 Procedure outncc (send_char)

 Sends specified character through the RS-232 interface.

****************************************************************************/

outncc (send_char)

char   send_char ;
{

/* Loop until the transmit hold register is empty */

  while ((inbyt (PORT_LSR) & 0x20) == 0)
    ;

  outbyt (PORT_THR, send_char) ;	/* send the byte */

  }    /* outncc() */



/****************************************************************************
 Function getncc () : char

 Receives a character from the RS-232 interface.

****************************************************************************/

getncc ()

{
  char	 ret_char ;
  char	 *ptr ;
  char	 *ptr2 ;

/* Loop until the head and tail pointer are different. */

  while (read_p == write_p)
    kbabort () ;

/* Can now unload the character received from the character spool. */

  ret_char = *read_p++ ;
  if (read_p > tail_p)
    read_p = head_p ;

  #asm
   cli
  #endasm
  if (read_p == write_p)
    data_avail = FALSE ;
  #asm
   sti
  #endasm

  return (ret_char) ;

  }    /* getncc() */



usage()
{
puts("usage: PMODEM baud S  FILENAME.EXT - Send a file from the host\n") ;
puts("       PMODEM baud SB FILENAME.EXT - Send a binary file from the host\n");
puts("       PMODEM baud R  FILENAME.EXT - Send a text file to the host\n");
puts("       PMODEM baud RB FILENAME.EXT - Send a binary file to the host\n");
puts("\n\nFor information on the Protocol used by this program read the");
puts(" file \"PMODEM.DQC\".\n");
exit(0);
}

index(s, t)

  char s[], t[] ;

{
  int	i, j, k ;

  for (i=0; s[i]!='\0'; i++) {
    for (j=i, k=0; t[k]!='\0' && s[j]==t[k]; j++,k++) ;
    if (t[k] == '\0') return(i) ;
  }
  return(ERROR) ;

}  /* index() */
     